@@ -1,12 +1,17 @@
 """skimpy provides summary statistics about variables in pandas data frames."""
+from __future__ import annotations  # This is here to get 'dict' typing for <3.10
+
+import datetime
+import os
 import re
 from collections import defaultdict
-from dataclasses import MISSING
 from itertools import chain
 from typing import Any
 from typing import Dict
 from typing import List
 from typing import Optional
+from typing import Tuple
+from typing import Union
 from unicodedata import normalize
 
 import numpy as np
@@ -14,7 +19,6 @@ import pandas as pd
 import rich
 from numpy.random import Generator
 from numpy.random import PCG64
-from rich import print
 from rich.columns import Columns
 from rich.console import Console
 from rich.panel import Panel
@@ -22,6 +26,21 @@ from rich.table import Table
 from rich.text import Text
 from typeguard import typechecked
 
+# the JSON # type: ignore comments throughout are sops to Python 3.8 issues
+
+# TypeAlias is only built-in for 3.10 and above
+try:
+    from typing import TypeAlias  # type: ignore
+except ImportError:
+    from typing_extensions import TypeAlias
+
+# polars check
+try:
+    import polars as pl
+except ImportError:
+    # no nothing
+    pass
+
 NULL_VALUES = {np.nan, "", None}
 
 CASE_STYLES = {
@@ -36,19 +55,23 @@ CASE_STYLES = {
     "upper",
 }
 
-console = Console()
-QUANTILES = [0, 0.25, 0.75, 1]
+QUANTILES = [0, 0.25, 0.5, 0.75, 1]
 HIST_BINS = 6
+# NB: unicode 1/2 and 8/8 blocks have inconsistent widths depending on font
+# systems, so, to make skimpy work on all systems, hist blocks are pinned to
+# nearest available consistent width block. This produces some distortion in
+# histogram but these are just visual indicators anyway and shouldn't be used
+# for deep inference.
 UNICODE_HIST = {
     0: " ",
     1 / 8: "▁",
     1 / 4: "▂",
     3 / 8: "▃",
-    1 / 2: "▄",
+    # 1 / 2: "▄",
     5 / 8: "▅",
     3 / 4: "▆",
     7 / 8: "▇",
-    1: "█",
+    # 1: "█",
 }
 # These are defined globally because they are used in more than one function
 DATE_COL_FIRST = "first"
@@ -57,15 +80,28 @@ NUM_COL_MEAN = "mean"
 COMPLETE_COL = "NA %"
 MISSING_COL = "NA"
 
+UNSUPPORTED_INFERRED_TYPES = [
+    "mixed-integer",
+    "mixed",
+    "mixed-integer-float",
+    "unknown-array",
+]
+
+try:
+    JSON: TypeAlias = dict[str, dict[str, Any]]  # type: ignore
+except TypeError:
+    JSON: TypeAlias = Dict[str, Dict[str, Any]]  # type: ignore
+
 
 @typechecked
 def _infer_datatypes(df: pd.DataFrame) -> pd.DataFrame:
     """Infers the, and applies new, datatypes of dataframe columns.
 
-    :param df: input dataframe of ambiguous col type
-    :type df: pd.DataFrame
-    :return: dataframe with column datatypes set to best of knowledge
-    :rtype: pd.DataFrame
+    Args:
+        df (pd.DataFrame): User data that we'd like to infer types on.
+
+    Returns:
+        pd.DataFrame: Same dataframe, but typed wherever possible.
     """
     df_types = (
         pd.DataFrame(df.apply(pd.api.types.infer_dtype, axis=0))
@@ -80,31 +116,38 @@ def _infer_datatypes(df: pd.DataFrame) -> pd.DataFrame:
             data_type = "int"
         elif col[1] == "floating":
             data_type = "float64"
+        elif col[1] == "timedelta64":
+            data_type = "timedelta64[ns]"
         elif col[1] == "datetime64":
             data_type = "datetime64[ns]"
         elif col[1] == "categorical":
             data_type = "category"
         elif col[1] == "boolean":
             data_type = "bool"
-        else:
-            data_type = col[1]
+        elif col[1] == "complex":
+            data_type = "complex128"
+        elif col[1] == "date":
+            # do nothing: "date" isn't supported, so we leave as an object.
+            # later on, we will still be able to pick up "date" using
+            # .select_dtypes, which is the important bit.
+            continue
+        # There is no else statement here because logic should never get to this point.
         df[col[0]] = df[col[0]].astype(data_type)
     return df
 
 
 @typechecked
-def _round_dataframe(df: pd.DataFrame) -> pd.DataFrame:
-    """Rounds dataframe to 2 s.f.
+def _round_series(s: pd.Series) -> pd.Series:
+    """Rounds numerical series to 2 s.f.
 
     Args:
-        df (pd.DataFrame): Input dataframe
+        s (pd.Series): Input series
 
     Returns:
-        pd.DataFrame: Dataframe with numbers rounded to 2 s.f.
+        pd.Series: Series with numbers rounded to 2 s.f.
     """
-    for col in df.select_dtypes("number"):
-        df[col] = df[col].apply(lambda x: float(f'{float(f"{x:.2g}"):g}'))
-    return df
+    s = s.apply(lambda x: float(f'{float(f"{x:.2g}"):g}'))
+    return s
 
 
 @typechecked
@@ -204,7 +247,7 @@ def _dataframe_to_rich_table(
     Returns:
         rich.table.Table: instance of Table from the rich package
     """
-    str_limit: int = 20  # For longer strings, limit chars shown.
+    COL_STR_LEN_LIMIT: int = 20  # For longer strings, limit chars shown.
     df = df.reset_index().rename(columns={"index": "column_name"})
     table = Table(show_footer=False, expand=True, title=table_name, show_header=True)
     # generate dict of types to colours
@@ -233,17 +276,17 @@ def _dataframe_to_rich_table(
         rows = _simplify_datetimes_in_array(rows)
     rows = [
         [
-            str(s).rstrip("0").rstrip(".") if ("." and type(s) == float) else s
+            str(s).rstrip("0").rstrip(".") if ("." and isinstance(s, float)) else s
             for s in row
         ]
         for row in rows
     ]
     for col in df.columns:
-        table.add_column(str(col), overflow="fold")
+        table.add_column(str(col), overflow="fold", max_width=COL_STR_LEN_LIMIT)
     for row in rows:
         row = [
             Text(
-                str(item)[:str_limit],
+                str(item),
                 style=pos_to_colour[i],
                 justify=pos_to_justification[i],
             )
@@ -274,7 +317,8 @@ def _create_unicode_hist(series: pd.Series) -> pd.Series:
 
     Given a pandas series of numerical values, returns a series with one
     entry, the original series name, and a histogram made up of unicode
-    characters.
+    characters. However, note that the histogram is very approximate, partly
+    due to limitations in how unicode is displayed across systems.
 
     Args:
         series (pd.Series): Numeric column of data frame for analysis
@@ -309,15 +353,15 @@ def _numeric_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     count_nans_vec = xf.isna().sum()
     data_dict = {
         MISSING_COL: count_nans_vec,
-        COMPLETE_COL: 100 * count_nans_vec / xf.shape[0],
-        NUM_COL_MEAN: xf.mean(),
-        "sd": xf.std(),
+        COMPLETE_COL: (100 * count_nans_vec / xf.shape[0]).round(2),
+        NUM_COL_MEAN: _round_series(xf.mean()),
+        "sd": _round_series(xf.std()),
     }
     display_quantiles_as_pct = 100
     quantiles_dict = dict(
         zip(
             ["p" + str(int(x * display_quantiles_as_pct)) for x in QUANTILES],
-            [xf.quantile(x) for x in QUANTILES],
+            [_round_series(xf.quantile(x)) for x in QUANTILES],
         )
     )
     data_dict.update(quantiles_dict)
@@ -345,7 +389,7 @@ def _category_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     count_nans_vec = xf.isna().sum()
     data_dict = {
         MISSING_COL: count_nans_vec,
-        COMPLETE_COL: 100 * count_nans_vec / xf.shape[0],
+        COMPLETE_COL: (100 * count_nans_vec / xf.shape[0]).round(2),
         "ordered": pd.Series(
             dict(zip(xf.columns, [xf[col].cat.ordered for col in xf.columns]))
         ),
@@ -370,7 +414,7 @@ def _bool_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     """
     data_dict = {
         "true": xf.sum(),
-        "true rate": xf.sum() / xf.shape[0],
+        "true rate": _round_series(xf.sum() / xf.shape[0]),
     }
     hist_series = pd.concat(
         [_create_unicode_hist(xf[col].dropna()) for col in xf.columns], axis=0
@@ -394,15 +438,17 @@ def _string_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     count_nans_vec = xf.isna().sum()
     data_dict = {
         MISSING_COL: count_nans_vec,
-        COMPLETE_COL: 100 * count_nans_vec / xf.shape[0],
-        "words per row": pd.Series(
-            dict(
-                zip(
-                    xf.columns,
-                    [
-                        xf[xf.columns[0]].str.count(" ").add(1).sum() / len(xf)
-                        for col in xf.columns
-                    ],
+        COMPLETE_COL: (100 * count_nans_vec / xf.shape[0]).round(2),
+        "words per row": _round_series(
+            pd.Series(
+                dict(
+                    zip(
+                        xf.columns,
+                        [
+                            xf[col].str.count(" ").add(1).sum() / len(xf)
+                            for col in xf.columns
+                        ],
+                    )
                 )
             )
         ),
@@ -410,18 +456,40 @@ def _string_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
             dict(
                 zip(
                     xf.columns,
-                    [
-                        xf[xf.columns[0]].str.count(" ").add(1).sum()
-                        for col in xf.columns
-                    ],
+                    [xf[col].str.count(" ").add(1).sum() for col in xf.columns],
                 )
-            )
+            ),
+            dtype="int",
         ),
     }
     summary_df = pd.DataFrame(data_dict)
     return summary_df
 
 
+@typechecked
+def _timedelta_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
+    """Summarise dataframe columns that have timedelta type. (NB not object type).
+
+    Args:
+        xf (pd.DataFrame):  Dataframe with columns of only timedelta types
+
+    Returns:
+        pd.DataFrame: A dataframe of summary statistics, with a number of rows
+        determined by number of columns of xf
+    """
+    count_nans_vec = xf.isna().sum()
+    # NB: timedelta doesn't play nicely with rounding
+    data_dict = {
+        MISSING_COL: count_nans_vec,
+        COMPLETE_COL: (100 * count_nans_vec / xf.shape[0]).round(2),
+        NUM_COL_MEAN: xf.mean().dt.floor("s"),
+        "median": xf.min().dt.floor("s"),
+        "max": xf.max().dt.ceil("s"),
+    }
+    summary_df = pd.DataFrame(data_dict)
+    return summary_df
+
+
 @typechecked
 def _datetime_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     """Summarise dataframe columns that have datetime type.
@@ -436,7 +504,7 @@ def _datetime_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
     count_nans_vec = xf.isna().sum()
     data_dict = {
         MISSING_COL: count_nans_vec,
-        COMPLETE_COL: 100 * count_nans_vec / xf.shape[0],
+        COMPLETE_COL: (100 * count_nans_vec / xf.shape[0]).round(2),
         DATE_COL_FIRST: pd.Series(
             dict(
                 zip(
@@ -464,35 +532,43 @@ def _datetime_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
 
 
 @typechecked
-def skim(
-    df_in: pd.DataFrame, header_style: str = "bold cyan", **colour_kwargs: str
-) -> None:
-    """Skim a data frame and return statistics.
-
-    skim is an alternative to pandas.DataFrame.summary(), quickly providing
-    an overview of a data frame. It produces a different set of summary
-    functions based on the types of columns in the dataframe. You may get
-    better results from ensuring that you set the datatypes in your dataframe
-    you want before running skim.
-    The colour_kwargs (str) are defined in _dataframe_to_rich_table.
+def _delete_unsupported_columns(df: pd.DataFrame) -> pd.DataFrame:
+    """This will remove the pd.api.types.infer_dtype types that are not
+    supported.
 
     Args:
-        df_in (pd.DataFrame): Dataframe to skim
-        header_style (str): A style to use for headers. See Rich API Styles.
-        colour_kwargs (dict[str]): colour keyword arguments for rich table
+        df (pd.DataFrame): _description_
 
-    Examples
-    --------
-    Skim a dataframe
+    Returns:
+        pd.DataFrame: _description_
+    """
+    df_types = (
+        pd.DataFrame(df.apply(pd.api.types.infer_dtype, axis=0))
+        .reset_index()
+        .rename(columns={"index": "column", 0: "type"})
+    )
+    df_types["delete"] = df_types["type"].isin(UNSUPPORTED_INFERRED_TYPES)
+    for _, row in df_types.iterrows():
+        if row["delete"]:
+            df = df.drop(row["column"], axis=1)
+    if df.empty:
+        raise ValueError(
+            f"Your input dataframe only has unsupported column types, eg {', '.join(UNSUPPORTED_INFERRED_TYPES)}"
+        )
+    return df
 
-    >>> df = pd.DataFrame(
-            {
-             'col1': ['Philip', 'Turanga', 'bob'],
-             'col2': [50, 100, 70],
-             'col3': [False, True, True]
-            })
-    >>> df["col1"] = df["col1"].astype("string")
-    >>> skim(df)
+
+@typechecked
+def _skim_computation(
+    df_in: pd.DataFrame,
+) -> Tuple[rich.table.Table, JSON]:
+    """Performs the under-the-hood summary statistics.
+
+    Args:
+        df_in (pd.DataFrame): Input pandas dataframe to create a summary of.
+
+    Returns:
+        [rich.table.Table, JSON]: Rich table grid to print to console, JSON of summary stats.
     """
     if hasattr(df_in, "name") and "name" not in df_in.columns:
         name = df_in.name
@@ -501,56 +577,77 @@ def skim(
 
     # Make a copy so as not to mess with dataframe
     df = df_in.copy()
+    # remove any columns with types that are not currently supported
+    df = _delete_unsupported_columns(df)
     # Perform inference of datatypes
     df = _infer_datatypes(df)
 
+    header_style = "bold cyan"  # fixed
+    # main data dict
+    json_data = {}
     # Data summary
     tab_1_data = {"Number of rows": df.shape[0], "Number of columns": df.shape[1]}
+    dat_sum_table_title = "Data Summary"
     dat_sum_table = Table(
-        title="Data Summary", show_header=True, header_style=header_style
+        title=dat_sum_table_title, show_header=True, header_style=header_style
     )
     dat_sum_table.add_column(name)
     dat_sum_table.add_column("Values")
     for key, val in tab_1_data.items():
         dat_sum_table.add_row(key, str(val))
-    # Data tpes
+    json_data.update({dat_sum_table_title: tab_1_data})
+    # Data types
+    data_types_title = "Data Types"
     types_sum_table = Table(
-        title="Data Types", show_header=True, header_style=header_style
+        title=data_types_title, show_header=True, header_style=header_style
     )
     tab_2_data = df.dtypes.astype(str).value_counts().to_dict()
     types_sum_table.add_column("Column Type")
     types_sum_table.add_column("Count")
     for key, val in tab_2_data.items():
         types_sum_table.add_row(key, str(val))
+    json_data.update({data_types_title: tab_2_data})
     # Categorys
     if "category" in df.dtypes.astype(str).to_list():
+        cat_section_title = "Categories"
         xf = pd.DataFrame(df.dtypes.astype(str))
         cat_sum_table = Table(
-            title="Categories", show_header=True, header_style=header_style
+            title=cat_section_title, show_header=True, header_style=header_style
         )
         header_string = f"[{header_style}]Categorical Variables[/{header_style}]"
         cat_sum_table.add_column(header_string)
         cat_names = list(xf[xf[0] == "category"].index)
         for cat in cat_names:
             cat_sum_table.add_row(cat)
+        json_data.update(
+            {cat_section_title: {"Columns": {cat_name for cat_name in cat_names}}}
+        )
     # Summaries of cols of specific types
     types_funcs_dict = {
         "number": _numeric_variable_summary_table,
         "category": _category_variable_summary_table,
-        "datetime": _datetime_variable_summary_table,
-        "string": _string_variable_summary_table,
         "bool": _bool_variable_summary_table,
+        "datetime": _datetime_variable_summary_table,
+        # Please note that datetime.date is not directly supported by pandas and registers as being of type "object"
+        datetime.date: _datetime_variable_summary_table,  # Re-use of fn intended.
+        "timedelta64[ns]": _timedelta_variable_summary_table,
+        "string": _string_variable_summary_table,
     }
     list_of_tabs = []
     for col_type, summary_func in types_funcs_dict.items():
-        xf = df.select_dtypes(col_type)
+        if col_type == "number":
+            # timedelta and datetime are technically integers, so exclude these
+            xf = df.select_dtypes(col_type, exclude=["datetime", "timedelta"])
+        else:
+            xf = df.select_dtypes(col_type)
         if not xf.empty:
             sum_df = summary_func(xf)
-            list_of_tabs.append(
-                _dataframe_to_rich_table(
-                    col_type, _round_dataframe(sum_df)  # , **colour_kwargs
-                )
-            )
+            # for rich tables, we need to stringify
+            # specialised and unsupported col types, such as datetime.date,
+            # that are actually registered as object type
+            col_type_to_rich = str(col_type)
+            list_of_tabs.append(_dataframe_to_rich_table(col_type_to_rich, sum_df))
+            json_data.update({col_type_to_rich: sum_df.to_dict()})
     # Put all of the info together
     grid = Table.grid(expand=True)
     tables_list = [dat_sum_table, types_sum_table]
@@ -562,9 +659,130 @@ def skim(
         grid.add_row(sum_tab)
     # Weirdly, iteration over list of tabs misses last entry
     grid.add_row(list_of_tabs[-1])
+    return grid, json_data
+
+
+@typechecked
+def skim(
+    df_in: Union[pd.DataFrame, pl.DataFrame],
+) -> None:
+    """Skim a pandas or polars dataframe and return visual summary statistics on it.
+
+    skim is an alternative to pandas.DataFrame.describe(), quickly providing
+    an overview of a data frame via a table displayed in the console. It produces a different set of summary
+    functions based on the types of columns in the dataframe. You may get
+    better results from ensuring that you set the datatypes in your dataframe
+    you want before running skim.
+
+    Note that any unknown column types, or mixed column types, will not be
+    processed.
+
+    Args:
+        df_in (Union[pd.DataFrame, pl.DataFrame]): Dataframe to skim.
+
+    Examples
+    --------
+    Skim a dataframe
+
+        >>> df = pd.DataFrame(
+                {
+                'col1': ['Philip', 'Turanga', 'bob'],
+                'col2': [50, 100, 70],
+                'col3': [False, True, True]
+                })
+        >>> df["col1"] = df["col1"].astype("string")
+        >>> skim(df)
+    """
+    df_out = _convert_to_pandas(df_in)
+    grid, json_data = _skim_computation(df_out)
+    console = Console(record=True)
     console.print(Panel(grid, title="skimpy summary", subtitle="End"))
 
 
+def _convert_to_pandas(df_in: Union[pd.DataFrame, pl.DataFrame]) -> pd.DataFrame:
+    if isinstance(df_in, pl.DataFrame):
+        df_out = df_in.to_pandas()
+    else:
+        df_out = df_in.copy()
+    return df_out
+
+
+@typechecked
+def skim_get_data(
+    df_in: Union[pd.DataFrame, pl.DataFrame],
+) -> Union[JSON, str]:
+    """Skim a pandas or polars dataframe and return summary statistics as a dictionary, and without printing to the console.
+
+    skim is an alternative to pandas.DataFrame.describe(), quickly providing
+    an overview of a data frame via a table of summary statistics. It produces a different set of summary
+    functions based on the types of columns in the dataframe. You may get
+    better results from ensuring that you set the datatypes in your dataframe
+    you want before running skim.
+
+    Note that any unknown column types, or mixed column types, will not be
+    processed.
+
+    Args:
+        df_in (Union[pd.DataFrame, pl.DataFrame]): Dataframe to get summary statistics on.
+
+    Returns:
+        Union[JSON, str]: Dictionary of summary statistics.
+    """
+    df_out = _convert_to_pandas(df_in)
+    _, json_data = _skim_computation(df_out)
+    return json_data
+
+
+@typechecked
+def skim_get_figure(
+    df_in: Union[pd.DataFrame, pl.DataFrame],
+    save_path: Union[os.PathLike, str],
+    format: str = "svg",
+) -> None:
+    """Skim a pandas or polars dataframe, print the stats to the console, and save a version of the table as an SVG, HTML, or text file.
+
+    skim is an alternative to pandas.DataFrame.describe(), quickly providing
+    an overview of a data frame via a table of summary statistics. It produces a different set of summary
+    functions based on the types of columns in the dataframe. You may get
+    better results from ensuring that you set the datatypes in your dataframe
+    you want before running skim.
+
+    Note that any unknown column types, or mixed column types, will not be
+    processed.
+
+    Args:
+        df_in (Union[pd.DataFrame, pl.DataFrame]): Dataframe to skim.
+        save_path (Union[os.PathLike, str]): Path to save figure to (include extension).
+        format (str, optional): svg, html, or text. Defaults to "svg".
+    """
+    df_out = _convert_to_pandas(df_in)
+    grid, _ = _skim_computation(df_out)
+    console = Console(record=True)
+    console.print(Panel(grid, title="skimpy summary", subtitle="End"))
+    if not isinstance(save_path, str):
+        save_path_str = str(save_path)
+    else:
+        save_path_str = save_path
+    # for when support is python >=3.10 only
+    # match format.lower():
+    #     case "svg":
+    #         console.save_svg(save_path_str)
+    #     case "html":
+    #         console.save_html(save_path_str)
+    #     case "text":
+    #         console.save_text(save_path_str)
+    #     case _:
+    #         raise ValueError("Format must be: svg, html, or text")
+    if format.lower() == "svg":
+        console.save_svg(save_path_str)
+    elif format.lower() == "html":
+        console.save_html(save_path_str)
+    elif format.lower() == "text":
+        console.save_text(save_path_str)
+    else:
+        raise ValueError("Format must be: svg, html, or text")
+
+
 @typechecked
 def clean_columns(
     df: pd.DataFrame,
@@ -572,21 +790,21 @@ def clean_columns(
     replace: Optional[Dict[str, str]] = None,
     remove_accents: bool = True,
 ) -> pd.DataFrame:
-    """Clean messy column names of a pandas dataframe.
+    """Clean messy column names in a pandas dataframe.
 
     Args:
         df (pd.DataFrame): Dataframe from which column names are to be cleaned.
         case (str, optional): The desired case style of the column name. Defaults to "snake".
 
-                - 'snake': 'column_name'
-                - 'kebab': 'column-name'
-                - 'camel': 'columnName'
-                - 'pascal': 'ColumnName'
-                - 'const': 'COLUMN_NAME'
-                - 'sentence': 'Column name'
-                - 'title': 'Column Name'
-                - 'lower': 'column name'
-                - 'upper': 'COLUMN NAME'
+                - 'snake' produces 'column_name';
+                - 'kebab' produces 'column-name';
+                - 'camel' produces 'columnName';
+                - 'pascal' produces 'ColumnName';
+                - 'const' produces 'COLUMN_NAME';
+                - 'sentence' produces 'Column name';
+                - 'title' produces 'Column Name';
+                - 'lower' produces 'column name';
+                - 'upper' produces 'COLUMN NAME';
 
         replace (Optional[Dict[str, str]], optional): Values to replace in the column names. Defaults to None.
 
@@ -605,26 +823,23 @@ def clean_columns(
     Clean column names by converting the names to camel case style, removing accents,
     and correcting a mispelling.
 
-    >>> df = pd.DataFrame(
-                          {
-                           'FirstNom': ['Philip', 'Turanga'],
-                           'lastName': ['Fry', 'Leela'],
-                           'Téléphone': ['555-234-5678', '(604) 111-2335']
-                          })
+        >>> df = pd.DataFrame(
+                            {
+                            'FirstNom': ['Philip', 'Turanga'],
+                            'lastName': ['Fry', 'Leela'],
+                            'Téléphone': ['555-234-5678', '(604) 111-2335']
+                            })
 
-    >>> clean_columns(df, case='camel', replace={'Nom': 'Name'})
-      firstName lastName       telephone
-    0    Philip      Fry    555-234-5678
-    1   Turanga    Leela  (604) 111-2335
+        >>> clean_columns(df, case='camel', replace={'Nom': 'Name'})
+        firstName lastName       telephone
+        0    Philip      Fry    555-234-5678
+        1   Turanga    Leela  (604) 111-2335
     """
     if case not in CASE_STYLES:
         raise ValueError(
             f"case {case} is invalid, options are: {', '.join(c for c in CASE_STYLES)}"
         )
 
-    # Store original column names for creating cleaning report
-    orig_columns = df.columns.astype(str).tolist()
-
     if replace:
         df = df.rename(columns=lambda col: _replace_values(col, replace))
 
@@ -633,12 +848,6 @@ def clean_columns(
 
     df = df.rename(columns=lambda col: _convert_case(col, case))
     df.columns = _rename_duplicates(df.columns, case)
-    # Count the number of changed column names
-    new_columns = df.columns.astype(str).tolist()
-    cleaned = [
-        1 if new_columns[i] != orig_columns[i] else 0 for i in range(len(orig_columns))
-    ]
-    print(f"{sum(cleaned)} column names have been cleaned")
     return df
 
 
@@ -740,7 +949,7 @@ def _remove_accents(name: Any) -> Any:
 
 
 @typechecked
-def _rename_duplicates(names: pd.Index, case: str) -> Any:
+def _rename_duplicates(names: pd.Index | list[str], case: str) -> Any:
     """Rename duplicated column names to append a number at the end."""
     if case in {"snake", "const"}:
         sep = "_"
@@ -765,7 +974,7 @@ def _rename_duplicates(names: pd.Index, case: str) -> Any:
 
 @typechecked
 def generate_test_data() -> pd.DataFrame:
-    """Generate dataframe with several different datatypes.
+    """Generate a pandas dataframe with several different datatypes.
 
     For testing skimpy, it's convenient to have a dataset with many different
     data types. This function creates that dataframe.
@@ -777,7 +986,7 @@ def generate_test_data() -> pd.DataFrame:
     --------
     Generate test data to demonstrate how skimpy works.
 
-    >>> df = generate_test_data()
+        >>> df = generate_test_data()
     """
     seed = 34729
     rng = Generator(PCG64(seed))
@@ -808,9 +1017,14 @@ def generate_test_data() -> pd.DataFrame:
     df.loc[[3, 5, 8, 9, 14, 22], "text"] = None
     df["text"] = df["text"].astype("string")
     # add a datetime column
-    df["date"] = pd.date_range("2018-01-01", periods=len_df, freq="M")
-    df["date_no_freq"] = rng.choice(
+    df["datetime"] = pd.date_range("2018-01-01", periods=len_df, freq="M")
+    df["datetime_no_freq"] = rng.choice(
         (pd.to_datetime(pd.Series(["01/01/2022", "03/04/2023", "01/05/1992"]))), len_df
     )
-    df.loc[[3, 12, 0], "date_no_freq"] = pd.NaT
+    df["datetime.date"] = df["datetime"].dt.date
+    df["datetime.date_no_freq"] = df["datetime_no_freq"].dt.date
+    df.loc[[3, 12, 0], "datetime_no_freq"] = pd.NaT
+    timedelta_array = rng.multinomial(40, [1 / 7] * 5, len_df).ravel()
+    df["time diff"] = pd.Series([pd.Timedelta(x, "d") for x in timedelta_array])
+    df.loc[[22, 1, 13, 65, 120], "time diff"] = pd.NaT
     return df
